//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OalServer.cc	--- implementation of OalServer classes
//
//	DESCRIPTION:
//		Implementation part of OalServer
//
//	CHANGE HISTORY:
//		UL, 1998-07-24: 1.0.0: - first version extracted from old OalAtoms.cc
//		UL, 1999-08-20: 1.1.0: - added diagnostic messages
//		UL, 1999-10-29: 1.2.0: - reset warp-around counter of
//								 InstrFrameCounter<> objects if ccd/node id
//								 has changed in state
//		UL, 1999-11-02: 1.3.0: - bail out in setCcdNr()/setNodeNr() on
//								 seeing invalid numbers
//		UL, 1999-12-17: 1.4.0: - implemented oalServer<<block/set
//								 functions
//		UL, 2000-04-13: 1.5.0: - made data mode list ODF ICD 2.4 compliant
//		UL, 2000-05-19: 1.6.0: - more diagnostic messages from setState()
//							   - prepending ODF directory name to data set
//								 name only of not found in current directory
//		UL, 2000-06-07: 1.7.0: - instantiate TimeCorrelator with much
//								 relaxed RMS criterion for OBT-MET fit
//		UL, 2000-11-09: 1.8.0: - several changes to fix SPR 1825: state
//								 variable dataMode is no longer obtained
//								 from contents of DATATYPE keyword but
//								 from asking the ODF for the data mode
//								 of the exposure with a given number
//		UL, 2000-12-07: 1.9.0: - OBT-MET fit parameters can be given in
//								 environment variable
//									SAS_OBT_MET_FIT
//		UL, 2000-12-08: 1.9.1: - a few trival changes to make the optimizer
//								 silent
//		UL, 2000-12-12: 1.9.2: - relaxed RMS criterion of OBT-MET fit to 1.-2
//
//	AUTHOR:
//		U. Lammers, Fri Jul 24 09:52:44 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/OalServer.cc,v 1.30 2001/04/02 11:46:04 ulammers Exp $
//

#include <fstream>

#include "OalServer.h"
#include "DalUtils.h"
#include "Tracer.h"
#include "errstr.h"
#include "stringutil.h"


OalServer &OalServer::instance()
{
	static OalServer *oalServerInstance = new OalServer;

	return *oalServerInstance;
} // oalServerFunc()


//
//	---( OalStateable )----------------------------------------
//
void OalStateable::instrument(Xmm::Instrument instrumentId) const
{
	PROC

	_instrument = instrumentId;

	ENDPROC
} // OalStateable::instrument()


void OalStateable::dataMode(Xmm::DataMode dataMode) const
{
	PROC

	_dataMode = dataMode;

	ENDPROC
} // OalStateable::dataMode()


void OalStateable::exposureIndex(unsigned exposureIndex) const
{
	PROC

	if (exposureIndex >= 1)
		_exposureIndex = exposureIndex;
	else
		errstr<<"Exposure index must be greater than 0"
			  <<Sas::error("IllegalExposureNumber");

	//	need to reset wraps around counter / reset time
	//	in InstrFrameCounterVectors
	InstrFrameCounterVector<EMOScam>::reset();
	InstrFrameCounterVector<EPNcam>::reset();
	InstrFrameCounterVector<RGScam>::reset();
	InstrFrameCounterVector<OMcam>::reset();

	//	define exposure number
	const Exposure *exp =
		OalServer::instance().odf()->exposure(_instrument, _dataMode,
											  _exposureIndex);
	_exposureNumber = exp->counter();

	ENDPROC
} // OalStateable::exposureIndex()


void OalStateable::exposureId(const string &expId) const
{
	PROC

	if (expId.length() != 4)
		errstr<<"Given exposure designator `"<<expId<<"' does not match "
				"expected pattern `S|U[0-9][0-9][0-9]'"
			  <<Sas::error("InvalidExposureDesignator");

	_scheduled = toupper(expId[0]) == 'S';
	_exposureNumber = toNumber<unsigned>(expId.substr(1, 3));
	Xmm::DataMode mode = oalServer.odf()->whatMode(_instrument,
												   _exposureNumber,
												   _scheduled);
	oalServer.dataMode(mode);
	const Exposure *exp = oalServer.odf()->exposure(_instrument, mode,
													_exposureNumber,
													Odf::EXP_ODF);
	oalServer.exposureIndex(exp->index()+1);

	ENDPROC
} // OalStateable::exposureId()


void OalStateable::ccd(Ccd ccd) const
{
	PROC

	if (ccd< 1 || ccd>XmmPayLoad::instrument(_instrument).nCcds())
		errstr<<"Cannot set CCD number to "<<ccd<<" - valid range is [1, "
			  <<XmmPayLoad::instrument(_instrument).nCcds()<<"]"
			  <<Sas::error("NoSuchCcd");

	_ccd = ccd;

	//	need to reset wraps around counter in InstrFrameCounterVectors
	InstrFrameCounterVector<EMOScam>::resetWraps();
	InstrFrameCounterVector<EPNcam>::resetWraps();
	InstrFrameCounterVector<RGScam>::resetWraps();
	InstrFrameCounterVector<OMcam>::resetWraps();

	ENDPROC
} // OalStateable::ccd()


void OalStateable::node(Node node) const
{
	PROC

	if (UNT(node)>XmmPayLoad::instrument(_instrument).nNodes())
		errstr<<"Cannot set node ID to "<<node<<" - valid range is [0, "
			  <<XmmPayLoad::instrument(_instrument).nNodes()-1<<"]"
			  <<Sas::error("NoSuchNode");

	_node = node;

	//	need to reset wraps around counter in InstrFrameCounterVectors
	InstrFrameCounterVector<EMOScam>::resetWraps();
	InstrFrameCounterVector<EPNcam>::resetWraps();
	InstrFrameCounterVector<RGScam>::resetWraps();
	InstrFrameCounterVector<OMcam>::resetWraps();

	ENDPROC
} // OalStateable::node()


void OalStateable::attitudeFromStr(bool attitudeFromStr) const
{
	PROC

	_attitudeFromStr = attitudeFromStr;

	ENDPROC
} // OalStateable::setAttitudeFromStr()


//
//	---( OalServer )----------------------------------------------------------
//
//	--- constructors ---
//
void OalServer::init() const
{
	PROC

	_odf = 0;
	_timeCorrelator = 0;
	_attitudeLocator = 0;
	_positionLocator = 0;

	ENDPROC
} // OalServer::init()


OalServer::OalServer(Xmm::Instrument instrument,
					 Xmm::DataMode dataMode,
					 unsigned exposureIndex,
					 const string &exposureId,
					 Ccd ccd,
					 Node node,
					 bool attitudeFromStr)
	: OalStateable(instrument, dataMode, exposureIndex, exposureId,
				   ccd, node, attitudeFromStr)
{
	PROC
	init();
	ENDPROC
} // OalServer::OalServer()


OalServer::OalServer(const OalStateable &oalStateable)
			: OalStateable(oalStateable)
{
	PROC
	init();
	ENDPROC
} // OalServer::OalServer()


void OalServer::setState(const string &scienceFileName)
{
	PROC

	*this<<scienceFileName;

	ENDPROC
} // OalServer::setState()


void OalServer::setStateFromSetName(const string &scienceSetName)
{
	PROC

	checkOdf();
	const OdfSet *odfSet = _odf->selectSet(scienceSetName);
	instrument(odfSet->instrument());
	dataMode(odfSet->dataMode());
	const Exposure *exp = odf()->exposure(instrument(), dataMode(),
										  odfSet->exposureNr(),
										  Odf::EXP_ODF);
	exposureIndex(exp->index()+1);
	ccd(odfSet->ccdNr());
	node(Node(odfSet->nodeNr()));

	ENDPROC
} // OalServer::setStateFromSetName()


void OalServer::checkOdf() const
{
	PROC

	if (!_odf)
		//	no ODF defined yet - instantiate one first
		newOdf("");		

	ENDPROC
} // OalServer::checkOdf()


const Odf *OalServer::newOdf(const string &dirName) const
{
	PROC

	delete _odf;
	delete _timeCorrelator;	_timeCorrelator = 0;
	delete _attitudeLocator; _attitudeLocator = 0;
	delete _positionLocator; _positionLocator = 0;

	_odf = new Odf(dirName, const_cast<OalServer *>(this));
//
//	The following generates too much diagnostic output for the errstr
//	to handle - disable for the time being
//	if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
//		errstr<<"Successfully openened ODF in directory `"<<dirName<<"': "
//			  <<*_odf
//			  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"Successfully openened ODF in directory `"<<dirName<<"'"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN _odf;
} // OalServer::newOdf()


const Odf *OalServer::odf() const
{
	PROC

	checkOdf();

	RETURN _odf;
} // OalServer::odf()


static const char *SAS_OBT_MET_FIT = "SAS_OBT_MET_FIT";
									//
									//	name of environment variable
									//	containing parameters controlling
									//	the OBT-MET fit process:
									//
									//		deg[,rms[,frac]]
									//
									//	with:
									//		deg	: maximum degree of
									//			  fit polynomial
									//		rms	: root-mean-square fit
									//			  abandoning citerion
									//		frac: maximum fraction of
									//			  data points to be removed
									//			  in finding good fit
									//

const TimeCorrelator *OalServer::newTimeCorrelator(const STime &timeBase) const
{
	PROC

	delete _timeCorrelator;

	TimeCorrelator::ConstructionMethod	method = TimeCorrelator::TCS_ONLY;
											//	use TCS data only
	unsigned	deg = 1;					//	linear fit
	double		rms = 1.e-2;				//	fairly relaxed rms
	double		frac= .9;					//	90% of points can be removed

	//
	//	check if parameters are to be obtained from env
	//
	char *specPtr;
	if ((specPtr = getenv(SAS_OBT_MET_FIT))) {
		vector<string> v(breakup(specPtr, ","));
		if (v.size() < 1 || v.size() > 4)
			errstr<<"Error in decoding `"<<specPtr<<"' - syntax is: "
					"method,[degree[,rms[,frac]]]"
				  <<Sas::error("FormatError");

		unsigned m = toNumber<unsigned>(v[0]);
		if (m>UNT(TimeCorrelator::TCS_RECALCULATE_DELAYS))
			errstr<<"Construction method must be 0 or 1 - got "<<m
				  <<Sas::error("InvalidTimeCorrelatorConstruction");
		method = TimeCorrelator::ConstructionMethod(m);

		if (v.size() > 1) {		
			deg = toNumber<unsigned>(v[1]);
			if (deg<=0)
				errstr<<"Degree of polynomal for OBT-MET fit must be >=1"
						" - got "<<deg<<Sas::error("InvalidPolyDegree");
		}

		if (v.size() > 2) {
			rms = toNumber<double>(v[2]);
			if (rms <= 0.)
				errstr<<"RMS of OBT-MET fit must be >=0 - got: "<<rms
					  <<Sas::error("InvalidRMSvalue");
		}

		if (v.size() > 3)
			frac = toNumber<double>(v[3]);
			if (frac < 0. || frac >= 1.)
				errstr<<"Fraction of removable points in OBT-MET fit must be "
						"in interval [0, 1) - got: "<<frac
					  <<Sas::error("InvalidFracvalue");
	}

	errstr<<"Constructing time correlator object with:\n"
		  <<"   construction method         : "<<method<<endl
		  <<"   maximum polynomial degree   : "<<deg<<endl
		  <<"   RMS of fit                  : "<<rms<<endl
		  <<"   fraction of removable points: "<<frac
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	TimeCorrelator::FitPolySpecs obtMetFitPolySpecs = { deg, rms, frac };
	TimeCorrelator::FitPolySpecs ertMetFitPolySpecs = { deg, rms, frac };

	RETURN _timeCorrelator = new TimeCorrelator(this,
												obtMetFitPolySpecs,
												ertMetFitPolySpecs,
												method,
												timeBase);
} // OalServer::newTimeCorrelator()


const TimeCorrelator *OalServer::timeCorrelator() const
{
	PROC

	RETURN _timeCorrelator ? _timeCorrelator : newTimeCorrelator();
} // OalServer::timeCorrelator()


const AttitudeLocator *OalServer::newAttitudeLocator(double maxDelta)
const
{
	PROC

	delete _attitudeLocator;
	_attitudeLocator = new AttitudeLocator(this, maxDelta);

	RETURN _attitudeLocator;
} // OalServer::newAttitudeLocator()



const AttitudeLocator *OalServer::attitudeLocator() const
{
	PROC

	RETURN _attitudeLocator ? _attitudeLocator : newAttitudeLocator(20.);
} // OalServer::attitudeLocator()


static const char *SAS_ORBIT = "SAS_ORBIT";
									//
									//	name of environment variable
									//	determining the type of orbit
									//	file that is used; possible value
									//	'R'	- reconstructed orbit
									//	'P' - predicted orbit
									//	anything else or undefined:
									//		- reconstructed or predicted
const XmmPositionLocator *OalServer::newPositionLocator(double maxDelta) const
{
	PROC

	delete _positionLocator;

	unsigned method = XmmPositionLocator::FROM_RECONSTRUCTED_ORBIT |
					  XmmPositionLocator::FROM_PREDICTED_ORBIT;

	//
	//	check if orbit determination switch from env
	//
	char *specPtr;
	if ((specPtr = getenv(SAS_ORBIT))) {
		string s(specPtr);
		if (s.length() > 0 && toupper(s[0]) == 'R')
			method = XmmPositionLocator::FROM_RECONSTRUCTED_ORBIT;
		if (s.length() > 0 && toupper(s[0]) == 'P')
			method = XmmPositionLocator::FROM_PREDICTED_ORBIT;
	}

	RETURN _positionLocator = new XmmPositionLocator(this, method, maxDelta);
} // OalServer::newPositionLocator()


const XmmPositionLocator *OalServer::positionLocator() const
{
	PROC

	RETURN _positionLocator ? _positionLocator : newPositionLocator();
} // OalServer::positionLocator()



const FrameCounterVector *OalServer::newFrameCounterVector(
			const int32 counterVec[], unsigned nCounter) const
{
	PROC

	FrameCounterVector *v = 0;
	switch (instrument()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2: v = new InstrFrameCounterVector<EMOScam>(
									counterVec, nCounter, this);
						 break;
		case Xmm::EPN  : v = new InstrFrameCounterVector<EPNcam>(
									counterVec, nCounter, this);
						 break;
		case Xmm::RGS1 :
		case Xmm::RGS2 : v = new InstrFrameCounterVector<RGScam>(
									counterVec, nCounter, this);
						 break;
		case Xmm::OM   : v = new InstrFrameCounterVector<OMcam>(
									counterVec, nCounter, this);
						 break;
		default:
						 errstr<<"Cannot allocate a frame counter vector "
								 "for instrument `"
							   <<Xmm::toString(instrument())<<"' - that does "
								 "not make much sense, does it?"
							   <<Sas::error("IllegalInstrument");
	}

	RETURN v;
} // OalServer::newFrameCounterVector()


OalServer::~OalServer()
{
	PROC

	delete _odf;
	delete _timeCorrelator;

	ENDPROC
} // OalServer::~OalServer()


OalServer &operator<<(OalServer &server, const string &dataSet)
{
	PROC

	string s(dataSet);
	//	see if the file can be opened 'as is'
	ifstream f(s.c_str());
	if (f)
		f.close();
	else
		//	prepend ODF directory name
		s = oalServer.odf()->directoryName() + '/' + s;

	DataSet *set = dataSetServer->open(s, DataSet::Read, DataSetServer::Low);

	server<<*set;

	dataSetServer->close(set);

	RETURN server;
} // operator<<


static ostream &statewarn(const string &listOfAttribs,
						  const string &statevar)
{
	PROC
	errstr<<"In defining the OAL state from a data set or block "
			"(one or more of) the mandatory attribute(s) "
		  <<listOfAttribs
		  <<" was/were not found - from this/these attribute(s) the actual "
		  <<statevar
		  <<" is derived which is now not possible - will "
			"assume a default value for the "<<statevar<<" - please check "
			"the diagnostic output below (requires verbosity >= Verbose) "
			"for the effective Oal state settings. Please note: "
			"In case of a possibly incorrect "<<statevar<<" all OAL "
			"operations which depend on the correct setting of this/these "
			"variable(s) will either fail to produce correct result or "
			"lead to an anomalous program termination. All OAL operations "
			"which do NOT depend on this/these variable(s) shall work "
			"normally."
		  <<Sas::warning("MandatoryAttributeNotFound");
	RETURN errstr;
} // statewarn()


OalServer &operator<<(OalServer &server, const vector<const Attributable *> av)
{
	PROC

	string modeStrings[] = { // EMOS
							 "IMAGE.EL", "TIME.EL", "RIMAGE.EL", "CTIME.EL",
							 "AUX.EL", "DIAG.IM", "OFFVAR.EL", "COUNT.EL",
							 "PERHK.EL", "HBR1_NPHK", "HBR2_NPHK",
							 "HBR3_NPHK", "EH_NPHK", "TH_NPHK", "BP_NPH",

							 // EPN
							 "BURST.EL", "OFFSET.IM", "NOISE.IM", "DISCLIN.IM",
							 "MAINHK.EL", "ADDHK.EL",

							 // ERM
							 "ERMCOUNT.EL", "SPECTRA.EL",

							 // RGS
							 "SPECT.EL", "HTR.EL", "FULLHK.EL", "TEMPHK.EL",
							 "DPP_HK.EL",

							 // OM
							 "FST.EL",
							 "ENG1.EL", "ENG2.EL", "ENG3.EL",
							 "ENG4.EL", "ENG5.EL", "ENG6.EL", "ENG7.EL",
							 "FAQ.EL", "WDW.EL", "BFAST.EL",
							 "TRH.EL", "REF.EL", "NONHK.EL",

							 //	SC
							 "ATT", "DUMATT", "PORB", "RECORB", "TIMECORR",
							 "HK_P1", "HK_P2", "HK_P3", "HK_P4", "HK_P5",
							 "HK_P6", "HK_P7", "HK_P8", "HK_P9", "SUM", "" };
	StringIntMap<Xmm::DataMode> dataMode(modeStrings);

	//
	//	make sure we have an ODF
	//
	oalServer.checkOdf();
	const Odf *odf = oalServer.odf();

	//
	//	see if we have an XMM data set
	//
	try {
		string s(DalUtils::attribute(av, "TELESCOP")->asString());
		if (s != "XMM")
			errstr<<"The data set used to define the Oal state does not seem "
					"to be an XMM set - the value of the attribute "
					"`TELESCOP\' is: "<<s
				  <<Sas::error("IllegalAttributeValue");
	} catch(...) {
		errstr<<"No attribute named `TELESCOP\' found"
			  <<Sas::error("MandatoryAttributeNotFound");
	}

	//
	//	determine the instrument
	//
	Xmm::Instrument instr = Xmm::SC;
	try {
		instr = Xmm::toInstrument(DalUtils::attribute(av,
													"INSTRUME")->asString());
	} catch(...) {
		errstr<<"The data set used to define the Oal state does not contain "
				"an attribute named `INSTRUME\' - this is needed to determine "
				"the current instrument."
			  <<Sas::error("MandatoryAttributeNotFound");
	}
	oalServer.instrument(instr);

	//
	//	determine exposure number/data mode
	//
	unsigned expNr = 1;
	try {
		string expId = DalUtils::attribute(av, "EXP_ID")->asString();
		if (expId.length() == 13)
			expNr = toNumber<unsigned>(expId.substr(10, 3));
		else
			errstr<<"Exposure identifier `"<<expId<<"' is invalid - "
					"state variable exposureNr will retain its default "
					"value"<<Sas::warning("IllegalExposureNumber");
	} catch(...) {
		statewarn("`EXP_ID\'", "exposure ID");
	}

	if (expNr > 0) {
		Xmm::DataMode mode = odf->whatMode(instr, expNr);
		if (mode == Xmm::DATA_MODE_NONE) {
			//
			//	try to get data mode from DATATYPE attribute
			//
			string dataType;
			try {
				dataType = DalUtils::attribute(av, "DATATYPE")->asString();
			} catch(...) {
				statewarn("`DATATYPE\'", "data mode");
			}
			Xmm::DataMode mode = dataMode[dataType];
			if (mode == Xmm::DATA_MODE_NONE)
				errstr<<"`"<<dataType<<"' is not a recognized data mode name -"
					  <<" datatype will retain its default setting"
					  <<Sas::warning("UnrecognizedValue");
		}
		oalServer.dataMode(mode);
		const Exposure *exp = oalServer.odf()->exposure(instr, mode,
														expNr, Odf::EXP_ODF);
		oalServer.exposureIndex(exp->index()+1);
	} else
		errstr<<"Exposure number must be greater than 0 - state variable "
				"expososurNr will retain its default value"
			  <<Sas::warning("IllegalExposureNumber");

	//
	//	determine CCD/node number
	//
	switch(instr) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
			try {
				oalServer.ccd(DalUtils::attribute(av, "CCDID")->asInt());
			} catch(...) {
				statewarn("`CCDID\'", "CCD");
			}
			try {
				oalServer.node(Xmm::toNode(DalUtils::attribute(av,
														"CCDNODE")->asInt()));
			} catch(...) {
				statewarn("`CCDNODE'", "node");
			}
			break;
		case Xmm::EPN:
			try {
				oalServer.ccd(EPNcam::toChipNr(
								DalUtils::attribute(av, "CCDID")->asInt(),
								DalUtils::attribute(av, "QUADRANT")->asInt()));
			} catch(...) {
				statewarn("`CCDID\', `QUADRANT\'", "CCD");
			}
			break;
		case Xmm::RGS1:
		case Xmm::RGS2:
			try {
				oalServer.ccd(DalUtils::attribute(av, "CCDID")->asInt());
			} catch (...) {
				statewarn("`CCDID\'", "CCD");
			}
			break;
		case Xmm::OM:
			oalServer.ccd(1);	//	lovely, there is only one CCD
			break;
		default:
			errstr<<"Initialization from `"<<Xmm::toString(instr)<<"' not "
					"supported"<<Sas::error("IllegalInstrument");
			break;
	}

	//
	//	all done, output distagnostic messges
	//
	errstr<<"OAL state set to:\n"
		  <<"   instrument             : "<<instr<<endl
		  <<"	data mode              : "<<oalServer.dataMode()<<endl
		  <<"   exposure index         : "<<oalServer.exposureIndex()<<endl
		  <<"   exposure number        : "<<oalServer.exposureNumber()<<endl
		  <<"   ccd number             : "<<oalServer.ccd()<<endl
          <<"   node number            : "<<oalServer.node()<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN server;
}


OalServer &operator<<(OalServer &server, const Block &block)
{
	PROC

	errstr<<"Setting OAL state from block `"<<block.qualifiedName()
		  <<"'"<<Sas::message(Msg::Lib, Msg::Verbose);
	vector<const Attributable *> av;
	av.push_back(&block);

	RETURN server<<av;
} // OalServer::operator<<


OalServer &operator<<(OalServer &server, const DataSet &dataSet)
{
	PROC

	errstr<<"Setting OAL state from data set `"<<dataSet.qualifiedName()
		  <<"'"<<Sas::message(Msg::Lib, Msg::Verbose);
	vector<const Attributable *> av;
	av.push_back(&dataSet);
	av.push_back(dataSet.block(0));

	RETURN server<<av;
} // OalServer::operator<<
