//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OalServer.h	--- access to Oal services
//
//	DESCRIPTION:
//		The defined class OalServer gives access to the services offered
//		by the ODF access layer
//
//	CHANGE HISTORY:
//		UL, 1997-11-25: 1.0.0: - skeleton
//		UL, 1998-07-24: 2.0.0: - overhaul
//
//	AUTHOR:
//		U. Lammers, Tue Nov 25 16:27:48 MET 1997
//
//	ACRONYMS:
//		AHF	: Attitude History File
//		OAL	: ODF Access Layer
//		OM	: Optical Monitor
//		THF	: Tracking History File
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/OalServer.h,v 1.12 2001/04/02 11:46:04 ulammers Exp $
///

#ifndef OAL_SERVER_H
#define OAL_SERVER_H

#include "TimeCorrelator.h"
#include "AttitudeLocator.h"
#include "XmmPositionLocator.h"
#include "TimeTags.h"
#include "Dal.h"
#include "Odf.h"
#include "Xmm.h"

#define oalServer	OalServer::instance()

class OalStateable {
public:
	OalStateable(Xmm::Instrument instrument = Xmm::EMOS1,
				 Xmm::DataMode dataMode = Xmm::IMAGING,
				 unsigned exposureIndex = 1,
				 const string &exposureId = "S001",
				 Ccd ccd = 1,
				 Node node = Xmm::PRIMARY_NODE,
				 bool attitudeFromStr = true) :
		_instrument(instrument),
		_dataMode(dataMode),
		_exposureIndex(exposureIndex), _exposureNumber(1), _scheduled(true),
		_exposureId(exposureId),
		_ccd(ccd),
		_node(node),
		_attitudeFromStr(attitudeFromStr) {}

	void instrument			(Xmm::Instrument instrumentId) const;
	void dataMode			(Xmm::DataMode dataMode) const;
	void exposureIndex		(unsigned exposureIndex) const;
	void exposureId			(const string &exposureId) const;
	void ccd				(Ccd ccd) const;
	void node				(Node node) const;
	void attitudeFromStr	(bool attitudeFromStr) const;

	Xmm::Instrument			instrument() const	{ PROC RETURN _instrument; }
	Xmm::DataMode			dataMode() const	{ PROC RETURN _dataMode; }
	unsigned				exposureIndex()const{ PROC RETURN _exposureIndex; }
	unsigned				exposureNumber()const{ PROC RETURN
															_exposureNumber; }
	bool					scheduled() const	{ PROC RETURN _scheduled; }
	Ccd						ccd() const			{ PROC RETURN _ccd; }
	Node					node() const		{ PROC RETURN _node; }
	bool					attitudeFromStr() const
											{ PROC RETURN _attitudeFromStr; }

	~OalStateable() {}

private:
	mutable Xmm::Instrument		_instrument;	// current instrument
	mutable Xmm::DataMode		_dataMode;		// current data mode
	mutable unsigned			_exposureIndex,	// current exposure index
								_exposureNumber;// current exposure number
	mutable bool				_scheduled;		// scheduled/unscheduled
	mutable string				_exposureId;	// "S001"
	mutable Ccd					_ccd;			// current CCD
	mutable	Node				_node;			// current readout node
	mutable bool				_attitudeFromStr;
												// get attitude data from
												// AHF or OM THF?
};


class OalServer : public OalStateable {
public:
	OalServer(Xmm::Instrument instrument = Xmm::EMOS1,
			  Xmm::DataMode dataMode = Xmm::IMAGING,
			  unsigned exposureIndex = 1,
			  const string &exposureId = "S001",
			  Ccd ccd = 1,
			  Node node = Xmm::PRIMARY_NODE,
			  bool attitudeFromStr = true);
	OalServer(const OalStateable &oalStateable);

	void setStateFromSetName(const string &scienceFileName);
	void setState(const string &odfFileName);
    friend OalServer &operator<<(OalServer &server, const string &dataSet);
    friend OalServer &operator<<(OalServer &server, const DataSet &dataSet);
    friend OalServer &operator<<(OalServer &server, const Block &block);
	friend OalServer &operator<<(OalServer &server,
										const vector<const Attributable *> av);
		//
		//	PURPOSE:
		//		Define the state and other components of this OalServer
		//		object from the contents of a Block in a Dal data set.
		//		In the former form all blocks of the specified data set
		//		are inspected for the informations which are needed to
		//		initialize the OalServer components (state)
		//		If the specified data set does not contain all the
		//		required information to re-initialize the object,
		//		corresponding warning message will be issue and the
		//		respective state variables retain their default values.
		//


	const Odf *newOdf(const string &dirName = "") const;
	const Odf *odf() const;

	const TimeCorrelator *timeCorrelator() const;
	const AttitudeLocator *attitudeLocator() const;
	const XmmPositionLocator *positionLocator() const;

	const FrameCounterVector *newFrameCounterVector(
			const int32 counterVec[], unsigned nCounter) const;
		//
		//	must be deleted after it's last use
		//

	static OalServer &instance();
		//
		//	return the single global instance of the OAL server
		//

	~OalServer();

private:
	mutable Odf				*_odf;			// the ODF
	mutable	TimeCorrelator	*_timeCorrelator;
											// time correlator object to
											// go from OBT/ERT to UTC
	mutable AttitudeLocator *_attitudeLocator;
											// locate S/C attitude from
											// AHF or OM THF data
	mutable XmmPositionLocator *_positionLocator;
											// locate S/C 

	//	methods
	void init() const;
	void checkOdf() const;
	void checkTimeCorrelator() const;

	const TimeCorrelator *newTimeCorrelator(
					const STime &timeBase = Xmm::missionReferenceTime) const;

	const AttitudeLocator *newAttitudeLocator(double maxDelta = 10.) const;

	const XmmPositionLocator *newPositionLocator(double maxDelta = 10.) const;
};

//
//	initialize OalServer from contents of Dal data set
//
OalServer &operator<<(OalServer &server, const string &dataSet);
OalServer &operator<<(OalServer &server, const DataSet &dataSet);
OalServer &operator<<(OalServer &server, const Block &block);

#endif
