//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Odf.h	--- utility classes related to Obervation Data File (ODF)
//
//	DESCRIPTION:
//		Various classes are defined in conjunction with needed functionality
//		to access information in an ODF.
//
//	CHANGE HISTORY:
//		UL, 1998-07-24: 1.0.0: - initial version
//		UL, 1998-09-30: 1.1.0: - new class with fields holding info
//								 matching Proposal Summary Record in
//								 ODF Summary File
//		UL, 2000-09-25: 1.2.0: - convenient access to IPPVs
//		UL, 2001-03-23: 1.3.0: - added:
//									* Odf::hasBadPixelMap()
//									* Odf::badPixelMap()
//		UL, 2001-04-27: 1.4.0: - fixed minor compilation problem under -O
//
//	AUTHOR:
//		U. Lammers, Fri Jul 24 09:46:32 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/Odf.h,v 1.24 2001/04/27 10:10:55 ulammers Exp $
///

#ifndef ODF_H
#define ODF_H

#include <vector>
#include <string>
#include <list>

#include "OalChild.h"
#include "StringIntMap.h"
#include "Xmm.h"
#include "STime.h"
#include "BadPixelMap.h"

class ExposureRecord {
public:
	enum ExposureType { SCIENCE, CALIBRATION };

	//	default constructor - does nothing
	ExposureRecord() {}

	//	construct from data tuple
	ExposureRecord(unsigned counter,
				   bool scheduledFlag,
				   ExposureType exposureType,
				   const STime &scheduledStartTime,
				   const STime &scheduledEndTime,
				   const STime &actualStartTime,
    			   const STime &actualEndTime);

	//	destructor
	~ExposureRecord() {}

	unsigned		counter() const { return _counter; }
	bool			scheduled() const { return _scheduled; }
	ExposureType	type() const { return _exposureType; }
	const STime		&scheduledStartTime() const {return _scheduledStartTime;}
	const STime		&scheduledEndTime() const { return _scheduledEndTime; }
	const STime		&actualStartTime() const {return _actualStartTime;}
	const STime		&actualEndTime() const { return _actualEndTime; }

private:
	//	data members
	unsigned		_counter;
	bool			_scheduled;
	ExposureType	_exposureType;
	STime			_scheduledStartTime, _scheduledEndTime,
					_actualStartTime, _actualEndTime;
};

ostream &operator<<(ostream &os, const ExposureRecord &expRec);


class OdfSet {
public:
	OdfSet();
	OdfSet(const string &basename, const string &dirname = ".");

	enum Type { TNONE = -1, EVENT_LIST, IMAGE, AUXILIARY, HOUSKEEPING,
				SPACECRAFT, SUMMARY };
	enum Format { FNONE = -1, COMPRESSED_FITS, COMPRESSED_ASCII, FITS, ASCII };

	//	string representation of instruments
	static StringIntMap<Xmm::Instrument> instruments;

	const string		name() const { return _dirname + '/' + _basename; }
	const string		&dirname() const { return _dirname; }
	const string		&basename() const { return _basename; }
	unsigned			revolutionNr() const { return _revolutionNr; }
	const string		&proposalId() const { return _proposalId; }
	Xmm::Instrument		instrument() const { return _instr; }
	bool				scheduled() const { return _scheduled; }
	unsigned			exposureNr() const { return _exposureNr; }
	unsigned		    ccdNr() const { return _ccdNr; }
	unsigned			nodeNr() const { return _nodeNr; }
	Xmm::DataMode		dataMode() const { return _dataMode; }
	OdfSet::Type		type() const { return _type; }
	OdfSet::Format		format() const { return _format; }
	bool				isValid() const { return _isValid; }

	friend ostream &operator<<(ostream &os, const OdfSet &fname);

	~OdfSet() {}

private:
	//	data members
	bool				_isValid;
	string				_basename,
						_dirname;
	unsigned			_revolutionNr;
	string				_proposalId;
	Xmm::Instrument		_instr;
	bool				_scheduled;
	unsigned			_exposureNr,
						_ccdNr,
						_nodeNr;
	Xmm::DataMode		_dataMode;
	OdfSet::Type		_type;
	OdfSet::Format		_format;
};


//
//	Instrument-Programmable-Parameter-Value
//
class IPPV : public string {
public:
	IPPV(const string &name, const string &value, bool invalid = false)
		: _name(name), _value(value), _invalid(invalid)
	{ PROC ENDPROC }	

	string name() const { PROC RETURN _name; }
	string value() const { PROC RETURN _value; }
	bool invalid() const { PROC RETURN _invalid; }
	template<class T>
	T value() { PROC RETURN !_invalid ? toNumber<T>(_value) : 0; }

	int valueAsInt() { PROC RETURN !_invalid ? toNumber<int>(_value) : 0; }
	double valueAsReal() { PROC RETURN !_invalid ? toNumber<double>(_value)
												 : 0.; }
	string valueAsString() { PROC RETURN value(); }

	~IPPV() { PROC ENDPROC }

private:
	string	_name, _value;
	bool	_invalid;
};


class Exposure : public ExposureRecord {
public:
	typedef map<string, string, less<string> >	IPPVmapType;

	Exposure() : ExposureRecord(), _index(0) {}
	Exposure(unsigned counter, bool scheduledFlag,
			 ExposureType exposureType,
			 const STime &scheduledStartTime,
			 const STime &scheduledEndTime,
			 const STime &actualStartTime,
			 const STime &actualEndTime,
			 const IPPVmapType &ippvMap,
			 unsigned index = 0) :
					ExposureRecord(counter, scheduledFlag, exposureType,
								   scheduledStartTime, scheduledEndTime,
								   actualStartTime, actualEndTime),
					_index(index), _ippvMap(ippvMap) {}

	bool hasIPPV(const string &name) const;
	IPPV getIPPV(const string &name) const;

	unsigned index() const { return _index; }
	void addSet(const OdfSet *odfSet);
	const list<const OdfSet *> &sets() const;

	~Exposure() {}

private:
	list<const OdfSet *>	_sets;
	unsigned				_index;
	IPPVmapType				_ippvMap;
};


class ProposalInfo {
public:
	ProposalInfo() {}
	~ProposalInfo() {}

	string	GoName, GoInstitute, GoAddress, GoEmail, AO, scienceType,
			targetName;
	double	targetRa, targetDec, obsDuration, boresightRa, boresightDec;
	string	posAngleConstLo, posAngleConstHi, posAngleOrigRef;
	short	emos1Prio, emos2Prio, epnPrio, rgs1Prio, rgs2Prio, omPrio;

	string GOcountry() const {
		vector<string> v(breakup(GoAddress, ","));
		string s(trim(v[v.size()-1], true));

		return s;	//	optimizer does not seem to like expressions in
					//	return statements
	}
};

ostream &operator<<(ostream &os, const ProposalInfo &proposalInfo);


class OalServer;	//	forward declaration
class Attributable;

class Odf : public virtual OalChild<OalServer> {
public:
	//	construct from directory name
	Odf(const string &directoryName = "", OalServer *oalEngine = 0);

	//	members
	string		directoryName() const { return _directoryName; }
	unsigned	orbitNr() const { return _orbitNr; }
	string		proposalId() const { return _proposalId; }
	string		observationId() const { return _observationId; }
	bool		isSlew() const { return _isSlew; }
	unsigned	sets() const { return _sets; }
	STime		observationStartTime() const
										{ return _observationStartTime; }
	STime		observationEndTime() const { return _observationEndTime; }
	const ProposalInfo
				&proposalInfo() const { return _proposalInfo; }

	unsigned	exposures(Xmm::Instrument instr, Xmm::DataMode dataMode)
				const;
	unsigned	sets(Xmm::Instrument, Xmm::DataMode dataMode,
					 unsigned ExposureNr) const;

	string		expandSetName(const string &setName) const;
	friend ostream &operator<<(ostream &os, const Odf &odf);

	enum ExposureNrType { EXP_ODF, EXP_OAL };
	const Exposure *exposure(Xmm::Instrument instr,
							 Xmm::DataMode dataMode,
							 unsigned n,
							 ExposureNrType expNrType = EXP_OAL,
							 bool lookForScheduledFlagMatch = false,
							 bool scheduled = true,
							 bool generateError = true) const;

	const Exposure *currentExposure() const;
	unsigned exposures() const;
	bool hasSet() const;
	const OdfSet *selectSet() const;	
	bool hasSet(Xmm::DataMode) const;
	const OdfSet *selectSet(Xmm::DataMode) const;
	bool hasSet(Xmm::Instrument, Xmm::DataMode) const;
	const OdfSet *selectSet(Xmm::Instrument, Xmm::DataMode) const;
	const OdfSet *selectSet(const string &setName) const;

	const OdfSet *selectSet(Xmm::Instrument instr, Xmm::DataMode dataMode,
							unsigned ExposureNr, unsigned ccdNr = 0,
							ExposureNrType expNrType = EXP_OAL,
							bool lookForScheduledFlagMatch = false,
							bool scheduled = true,
							bool generateError = true) const;

	//
	//	bad pixel retrieval
	//
	bool hasBadPixelMap() const;
	BadPixelMap badPixelMap() const;

	//	destructor
	~Odf();

private:
	//	data members
	string		_directoryName;
	typedef pair<Xmm::Instrument, Xmm::DataMode> InstrumentModePair;
	typedef map<InstrumentModePair, vector<Exposure *>,
									less<InstrumentModePair> > MapType;
	MapType		_exposures;

	bool		_isSlew;			// is this a slew ODF?
	unsigned	_orbitNr,			// orbit number
				_sets;				// number of set in this ODF
	string		_proposalId,		// proposal number
				_observationId;		// observation number
	STime		_observationStartTime,
				_observationEndTime;
									// observation start/end time
	list<OdfSet>	_setNames;
	ProposalInfo	_proposalInfo;
									// information from proposal database

	//	methods
	void registerOdfSet(OdfSet &odfSet);
	MapType::const_iterator findExposure(Xmm::Instrument,
										 Xmm::DataMode) const;
	Xmm::DataMode whatMode(Xmm::Instrument instr, unsigned exposureNr,
						   bool scheduled) const;
	Xmm::DataMode whatMode(Xmm::Instrument instr, unsigned exposureNr) const;
	friend OalServer &operator<<(OalServer &,
								 const vector<const Attributable *>);
	friend class OalStateable;
}; // Odf;

#endif
