//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrelator.cc	--- implementation of TimeCorrelator class
//
//	DESCRIPTION:
//		Implementation part of TimeCorrelator class
//
//	CHANGE HISTORY:
//		UL, 1998-07-16: 1.0.0: - first version
//		UL, 1999-03-04: 1.1.0: - necessary changes related to lowmem Dal
//		UL, 1999-04-28: 1.2.0: - changed fatal -> error
//		UL, 1999-08-20: 1.3.0: - added more diagnostic messages
//		UL, 2000-01-11: 1.4.0: - brought in line with ODF ICD issue 2.4
//								 (FTFINE column of TCS is now unsigned)
//		UL, 2001-03-02: 1.5.0: - implemented recalculation of signal
//								 propagation delays (using column ERTIMRAW
//								 in TCS); option to switch to either scheme
//								 via environment variable SAS_OBT_MET_FIT
//							   - bug fix: TCS file was not closed
//		UL, 2001-03-30: 1.3.0: - added
//									TimeCorrelator::toOnBoardTime()
//							   - give some routines better names
//		UL, 2001-04-04: 1.4.0: - construct MET-OBT fit polynomial
//								 with original OBT-MET poly specs
//		UL, 2001-05-31: 1.4.1: - ignore entries with UTCDAY==0 in TCS
//		UL, 2001-07-24: 1.5.0: - ignore also entries with UTCFLAG!=0 and
//								 all entries after an apparent OBT reset
//								 (see comments below); fixes problem of
//								 SPR 2127
//
//	AUTHOR:
//		UL, Thu Jul 16 10:40:31 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/TimeCorrelator.cc,v 1.30 2001/07/24 15:22:31 ulammers Exp $
//

#include "TimeCorrelator.h"
#include "CalOalUtility.h"
#include "OalServer.h"
#include "OalDalUtils.h"
#include "PhysicalConstants.h"
#include "Dal.h"
#include "Tracer.h"
#include "slatec.h"
#include "errstr.h"

#include <iomanip>
#include <algorithm>

using namespace CalOalUtility;

const FitPolynomial *TimeCorrelator::doPolyFit(FitPolySpecs &fitPolySpecs,
											   vector<double> &x,
											   vector<double> &y) const
{
	PROC

	#define MESSAGE Sas::message(Msg::Lib, Msg::Verbose);

	unsigned nPoints = x.size(), pointsRemoved = 0;
	FitPolynomial *fitPoly = new FitPolynomial(x, y, fitPolySpecs.degree,
											   fitPolySpecs.rms);
	double removedFrac = 0.;
	while (x.size() > fitPolySpecs.degree &&
		   removedFrac <= fitPolySpecs.discardPoints &&
		   fitPolySpecs.rms < fitPoly->rms()) {
		//
		//	no satisfactory fit found yet - delete fit polynomial,
		//	remove "worst" data point and re-calculate
		//
		vector<double> diff(x.size());
		for (unsigned n=0; n<x.size(); ++n)
			diff[n] = fabs(y[n] - fitPoly->evaluate(x[n]));
		vector<double>::iterator pos = max_element(diff.begin(), diff.end());
		x.erase(pos-diff.begin()+x.begin());
		y.erase(pos-diff.begin()+y.begin());
		removedFrac = DBL(++pointsRemoved)/DBL(nPoints);

		delete fitPoly;
		fitPoly = new FitPolynomial(x, y,fitPolySpecs.degree,fitPolySpecs.rms);
	}

	if (removedFrac > fitPolySpecs.discardPoints)
		errstr<<"No satisfactory fit found after removing "
			  <<setprecision(2)<<100.*removedFrac<<"% of the data points"
			  <<Sas::error("UnsuccessfulFit");

	if (x.size() <= fitPolySpecs.degree || fitPolySpecs.rms < fitPoly->rms())
		errstr<<"Could not find a satisfactory fit"
			  <<Sas::error("UnsuccessfulFit");

	//
	//	be informative
	//
	vector<double> diff(x.size());
	for (unsigned n=0; n<x.size(); ++n)
		diff[n] = y[n] - fitPoly->evaluate(x[n]);

	errstr<<"Fit polynomial found - degree: "<<fitPoly->degree()
		  <<" RMS: "<<setprecision(5)<<fitPoly->rms()
		  <<MESSAGE;
	errstr<<"Differences between data points and fit polynomial:";
	for (unsigned n=0; n<diff.size(); ++n)
		errstr<<"   "<<n+1<<" "<<setprecision(8)<<diff[n]
			  <<MESSAGE;

	fitPolySpecs.degree = fitPoly->degree();
	fitPolySpecs.rms = fitPoly->rms();
	fitPolySpecs.discardPoints = removedFrac;

	RETURN fitPoly;
	
	#undef MESSAGE
} // TimeCorrelator::doPolyFit()


void TimeCorrelator::construct(const string &timeCorrelationFileName,
							   ConstructionMethod cm)
{
	PROC

	if (cm == TCS_ONLY)
		constructFromTCSonly(timeCorrelationFileName);
	else if (cm == TCS_RECALCULATE_DELAYS)
		constructFromTCSrecalcDelays(timeCorrelationFileName);
	else
		errstr<<"Unknown TimeCorrelator construction method"
			  <<Sas::error("ThisShouldNotHappen");

	ENDPROC
} // TimeCorrelator::construct()


void TimeCorrelator::constructFromTCSonly(const string
													&timeCorrelationFileName)
{
	PROC

	errstr<<"Opening Time Correlation file `"<<timeCorrelationFileName
		  <<"' ..."<<Sas::message(Msg::Lib, Msg::Verbose);
	DataSet *set = dataSetServer->open(timeCorrelationFileName, DataSet::Read);
	Table *table = set->table(0);
	unsigned nRows = table->rows();

	ColumnData *erttimecol = table->column("ERTIME")->data(),
			   *cdobtcoarse	= table->column("OBTCOARS")->data(),
			   *cdobtfine = table->column("OBTFINE")->data(),
			   *cdutcday = table->column("UTCDAY")->data(),
			   *cdutcmilli = table->column("UTCMILLI")->data(),
			   *cdutcmicro = table->column("UTCMICRO")->data(),
			   *cdutcflag = table->column("UTCFLAG")->data();

	int32	*obtcoarse	= cdobtcoarse->int32Data();
	uint16	*obtfine	= cdobtfine->uint16Data();
	int16	*utcday		= cdutcday->int16Data();
	int32	*utcmilli	= cdutcmilli->int32Data();
	int16	*utcmicro	= cdutcmicro->int16Data(),
			*utcflag	= cdutcflag->int16Data();

	vector<double>	obt, met, ert;

	//
	//	get very first UTC field and establish _utcBaseTime
	//
	_utcBaseTime = STime(1958, 1, 1+utcday[0]);
	_utcOffset = _utcBaseTime - _baseTime;

	errstr<<"TimeCorrelator::utcBaseTime = "<<_utcBaseTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"TimeCorrelator::baseTime    = "<<_baseTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"TimeCorrelator::utcOffset   = "<<setprecision(12)
		  <<_utcOffset
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned lastObtCoarse = OalDalUtils::unsigned32(obtcoarse[0]),
			 nWraps = 0;
	for (unsigned n=0; n<nRows; ++n) {
		if (utcday[n] <= 0 || utcflag[n])
			continue;	//	ignore entries with
						//		* UTCDAY==0
						//		* UTCFLAG!=0
						//	note: some early ODFs e.g. 0078/0124100101 have
						//	UTCDAY==0 but UTCFLAG==0 so, UTCFLAG is not
						//	a reliable quality indicator

		//
		//	get OBT value
		//
		unsigned uobtcoarse = OalDalUtils::unsigned32(obtcoarse[n]);
		double diff = INT(uobtcoarse) - INT(lastObtCoarse);
		if (diff < 0.)
			if (-diff > .9*65536.*65536.)
				//
				//	coarse time must have had a wrap around
				//
				++nWraps;
			else {
				//
				//	this is an OBT reset (e.g. ODF 102_0099020301 jump
				//	from 1114402 to 33 at TCS record 2827) - cannot use
				//	subsequent time	correlation entries; ignore the rest
				//
				errstr<<"Detected apparent reset of onboard clock in time "
						"correlation data table `"<<table->qualifiedName()
					  <<" between records #"<<n<<'/'<<n+1
					  <<" where OBTCOARSE jumps from "
					  <<lastObtCoarse<<" to "<<uobtcoarse<<" - ignoring all "
						"subsequent records. The impact of this on "
						"the accuracy of event arrival times is very likely "
						"to be negligible. If you intend to perform "
						"high-accuracy timing analysis on this ODF data set, "
						"please consult the XMM-Newton helpdesk and ask for "
						"confirmation of this assessment. In doing so please "
						"detail the complete ODF designation and the version "
						"of the SAS you are using at this moment."
					  <<Sas::warning("OBTreset");
				break;
			}

		lastObtCoarse = uobtcoarse;
		obt.push_back(DBL(uobtcoarse) +
						 DBL(OalDalUtils::unsigned32(obtfine[n]))/65536. +
						 DBL(nWraps) * 65536. * 65536.);

		//
		//	get UTC value
		//
		met.push_back(STime(1958, 1, 1+utcday[n], 0, 0, 0,
						 1.e-3 * DBL(utcmilli[n]) + 1.e-6 * DBL(utcmicro[n])) -
					_utcBaseTime);

		//
		//	get ERT value
		//
		ert.push_back(STime(stringColumnRow(erttimecol, n)) - _utcBaseTime);

		errstr<<setprecision(12)
			  <<n<<" "<<obt[n]-obt[0]<<" "<<met[n]-met[0]<<" "<<ert[n]-ert[0]
			  <<Sas::message(Msg::Lib, Msg::Verbose);
	}

	//
	//	calculate fit polynomials
	//
	vector<double> met2(met);
	FitPolySpecs obtMetFitPolySpecs = _obtMetFitPolySpecs;
	_obtMetPoly = doPolyFit(_obtMetFitPolySpecs, obt, met);
	_metObtPoly = doPolyFit(obtMetFitPolySpecs, met, obt);
	_ertMetPoly = doPolyFit(_ertMetFitPolySpecs, ert, met2);

	errstr<<"OBT-MET fit polynomials successfully calculated"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	//	tidy up
	delete erttimecol;
	delete cdobtcoarse;
	delete cdobtfine;
	delete cdutcday;
	delete cdutcmilli;
	delete cdutcmicro;
	delete cdutcflag;

	dataSetServer->close(set);

	ENDPROC
} // TimeCorrelator::constructFromTCSonly()


void TimeCorrelator::constructFromTCSrecalcDelays(const string
													&timeCorrelationFileName)
{
	PROC

	errstr<<"Opening Time Correlation file `"<<timeCorrelationFileName
		  <<"' ..."<<Sas::message(Msg::Lib, Msg::Verbose);
	DataSet *set = dataSetServer->open(timeCorrelationFileName, DataSet::Read);
	Table *table = set->table(0);

	//
	//	check if we have column ERTIMRAW - if not this is an old-style
	//	TCS produced before Dec 2000
	//
	if (!table->hasColumn("ERTIMRAW")) {
		errstr<<"Table `"<<table->qualifiedName()<<"' has no column ERTIMRAW "
				"which is needed for the recalculation of time correlation "
				"data - will continue with XMCS generated values"
			  <<Sas::warning("NoSuchColumn");
		dataSetServer->close(set);
		constructFromTCSonly(timeCorrelationFileName);
	}

	vector<unsigned> obtcoarse, obtfine;
	CalOalUtility::fillVector(table, "OBTCOARS", obtcoarse);
	CalOalUtility::fillVector(table, "OBTFINE", obtfine);

	vector<string> erttime, gsid, erttimeraw;
	CalOalUtility::fillVector(table, "ERTIME", erttime);
	CalOalUtility::fillVector(table, "ERTIMRAW", erttimeraw);
	CalOalUtility::fillVector(table, "GSID", gsid);

	//vector<unsigned> utcday, utcmilli, utcmicro;
	//CalOalUtility::fillVector(table, "UTCDAY", utcday);
	//CalOalUtility::fillVector(table, "UTCMILLI", utcmilli);
	//CalOalUtility::fillVector(table, "UTCMICRO", utcmicro);

	const PositionLocator *posloc = oalServer.positionLocator();

	//
	//	get very first UTC field and establish _utcBaseTime
	//
	_utcBaseTime = erttimeraw[0];
	_utcOffset = _utcBaseTime - _baseTime;

	errstr<<"TimeCorrelator::utcBaseTime = "<<_utcBaseTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"TimeCorrelator::baseTime    = "<<_baseTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"TimeCorrelator::utcOffset   = "<<setprecision(12)
		  <<_utcOffset
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned N = obtcoarse.size();
	vector<double>	obt(N), met(N), ert(N);

	unsigned lastObtCoarse = obtcoarse[0],
			 nWraps = 0;
	for (unsigned n=0; n<N; ++n) {
		//
		//	get OBT value
		//
		if (obtcoarse[n] < lastObtCoarse)
			//
			//	coarse time must have had a wrap around
			//
			++nWraps;

		lastObtCoarse = obtcoarse[n];
		obt[n] = DBL(obtcoarse[n]) +
				 DBL(obtfine[n])/65536. + DBL(nWraps) * 65536. * 65536.;


		//double d1 = STime(1958, 1, 1+utcday[n], 0, 0, 0,
		//				 1.e-3 * DBL(utcmilli[n]) + 1.e-6 * DBL(utcmicro[n]))-
		//			STime(erttime[n]),
		//	   d2 = STime(erttimeraw[n])-STime(erttime[n]);
		//
		//	cerr<<obt[n]<<' '<<setprecision(10)<<d1<<' '<<d2<<endl;)

		//
		//	calculate MET value
		//
		unsigned short gsi = toNumber<unsigned short>(gsid[n]);
		const XmmGroundStation &gs = XmmMission::groundStation(gsi);
		STime t(erttimeraw[n]);
		CartesianVector xmmPos = posloc->stateVector(t).position,
						gsPos = gs.stateVector(t).position;
		double xmmGSdist = (gsPos - xmmPos).norm() * 1000.,	//	km -> m
			   propDelay =
					//	OBDH delay
					XmmMission::dataHandler().timingDelay() +

					//	signal travel time - NOTE: atmospheric propagation
					//	delays are of the order of 0.1 us (see email
					//	F. Dreger dated 2001-02-28), so, negligible
				    xmmGSdist/PhysicalConstants::c +

					// Ground station delays
					gs.timingDelay();

//		errstr<<"Total propagation delay at T="<<t.toString(6)
//			  <<": "<<propDelay<<"s"
//			  <<Sas::message(Msg::Lib, Msg::Verbose);

		met[n] = t - _utcBaseTime;
		met[n] -= propDelay;

		//
		//	get ERT value
		//
		ert[n] = STime(erttime[n]) - _utcBaseTime;

		errstr<<setprecision(12)
			  <<n<<" "<<obt[n]-obt[0]<<" "<<met[n]-met[0]<<" "<<ert[n]-ert[0]
			  <<Sas::message(Msg::Lib, Msg::Verbose);
	}

	//
	//	calculate fit polynomials
	//
	vector<double> met2(met);
	_obtMetPoly = doPolyFit(_obtMetFitPolySpecs, obt, met);
	_metObtPoly = doPolyFit(_obtMetFitPolySpecs, met, obt);
	_ertMetPoly = doPolyFit(_ertMetFitPolySpecs, ert, met2);

	errstr<<"OBT-MET fit polynomials successfully calculated"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	dataSetServer->close(set);

	ENDPROC
} // TimeCorrelator::constructFromTCSrecalcDelays()


TimeCorrelator::TimeCorrelator(const STime &timeBase)
	: OalChild<OalServer>(0), _baseTime(timeBase), _lastObtTag(0.),
							  _obtWraps(0), _obtMetPoly(0), _ertMetPoly(0)
{
	PROC

	_utcBaseTime = STime::earliest;

	_obtMetFitPolySpecs.degree = 0;
	_obtMetFitPolySpecs.rms = 0.;
	_obtMetFitPolySpecs.discardPoints = 0;

	_ertMetFitPolySpecs.degree = 0;
	_ertMetFitPolySpecs.rms = 0.;
	_ertMetFitPolySpecs.discardPoints = 0;

	ENDPROC
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(const OdfSet *timeCorrelationSet,
						   const FitPolySpecs &obtMetFitPolySpecs,
						   const FitPolySpecs &ertMetFitPolySpecs,
						   ConstructionMethod cm,
						   const STime &timeBase)
	: OalChild<OalServer>(0), _obtMetFitPolySpecs(obtMetFitPolySpecs),
	  _ertMetFitPolySpecs(ertMetFitPolySpecs),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{
	PROC
	construct(timeCorrelationSet->name(), cm);
	ENDPROC
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(const string &timeCorrelationSetName,
						   const FitPolySpecs &obtMetFitPolySpecs,
						   const FitPolySpecs &ertMetFitPolySpecs,
						   ConstructionMethod cm,
						   const STime &timeBase)
	: OalChild<OalServer>(0), _obtMetFitPolySpecs(obtMetFitPolySpecs),
	  _ertMetFitPolySpecs(ertMetFitPolySpecs),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{
	PROC
	construct(timeCorrelationSetName, cm);
	ENDPROC
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(const OalServer *oalEngine,
					const FitPolySpecs &obtMetFitPolySpecs,
					const FitPolySpecs &ertMetFitPolySpecs,
					ConstructionMethod cm,
					const STime &timeBase)
	: OalChild<OalServer>(oalEngine), _obtMetFitPolySpecs(obtMetFitPolySpecs),
	  _ertMetFitPolySpecs(ertMetFitPolySpecs),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{
	PROC

	assert(oalEngine);
	construct(oalEngine->odf()->selectSet(Xmm::SC, Xmm::TIME_CORRELATION, 1)
				->name(), cm);

	ENDPROC
} // TimeCorrelator::TimeCorrelator()


const TimeCorrelator::FitPolySpecs &TimeCorrelator::obtMetFitPolySpecs() const
{
	PROC
	RETURN _obtMetFitPolySpecs;
} // TimeCorrelator::obtMetFitPolySpecs()


const TimeCorrelator::FitPolySpecs &TimeCorrelator::ertMetFitPolySpecs() const
{
	PROC
	RETURN _ertMetFitPolySpecs;
} // TimeCorrelator::obtMetFitPolySpecs()


void TimeCorrelator::setBaseTime(const STime &) const
{
	PROC

	errstr<<"TimeCorrelator::setBaseTime is not yet implemented"
		  <<Sas::error("NotYetImplemented");

	ENDPROC
} // TimeCorrelator::setBaseTime()


double TimeCorrelator::toMissionEllapsedTime(double obtTag) const
{
	PROC
	if (obtTag < _lastObtTag &&			//	a time decrease
		_lastObtTag - obtTag > 320000.)	//	and difference is large enough
										//	to believe a wrap-around has
										//	occured
		++_obtWraps;

	_lastObtTag = obtTag;
	obtTag += DBL(_obtWraps) * 65536. * 65536.;

	RETURN _obtMetPoly->evaluate(obtTag) + _utcOffset;
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toMissionEllapsedTime(unsigned obtCoarse,
											 unsigned short obtFine) const
{
	PROC
	RETURN toMissionEllapsedTime(DBL(obtCoarse) + DBL(obtFine)/65536.);
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toOnBoardTime(double missionEllapsedTime) const
{
	PROC
	RETURN _metObtPoly->evaluate(missionEllapsedTime - _utcOffset);
} // TimeCorrelator::toOnBoardTime()


void TimeCorrelator::resetObtMetRelation() const
{
	PROC

	_lastObtTag = 0.;
	_obtWraps = 0;

	ENDPROC
} // TimeCorrelator::resetObtMetRelation()


void TimeCorrelator::plotObtMetRelation() const
{
	PROC

	errstr<<"TimeCorrelator::plotObtMetRelation is not yet implemented"
		  <<Sas::error("NotYetImplemented");

	ENDPROC
} // TimeCorrelator::plotObtMetRelation()


double TimeCorrelator::toMissionEllapsedTime(const string &ertString) const
{
	PROC
	RETURN toMissionEllapsedTime(STime(ertString));
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toMissionEllapsedTime(const STime &ertTag) const
{
	PROC
	double arg = ertTag - _utcBaseTime;

	RETURN _ertMetPoly->evaluate(arg) + _utcOffset;
} // TimeCorrelator::toMissionEllapsedTime()


void TimeCorrelator::plotErtMetRelation() const
{
	PROC

	errstr<<"TimeCorrelator::plotErtMetRelation is not yet implemented"
		  <<Sas::error("Not yet implemented");

	ENDPROC
} // TimeCorrelator::plotErtMetRelation()


TimeCorrelator::~TimeCorrelator()
{
	PROC

	delete _obtMetPoly;
	delete _ertMetPoly;

	ENDPROC
} // TimeCorrelator::~TimeCorrelator()



//
//	---------------------( FitPolynomial )-----------------------------------
//
FitPolynomial::FitPolynomial(const vector<double> &x,
							 const vector<double> &y,
							 const unsigned maxDegree, double rms)
	: _x(x), _y(y)
{
	PROC

	if (rms <= 0. || maxDegree == 0)
		errstr<<"RMS must be >= 0 and maximum allowed degree of fit polynomial"
				" must be >0"<<Sas::error("IllegalValues");

	unsigned nPoints = x.size();
	if (nPoints >= 2) {
		double	*weights = new double [nPoints],
				*vals = new double [nPoints],
				*xp = new double [nPoints],
				*yp = new double [nPoints];
		for (unsigned n=0; n<nPoints; ++n) {
			xp[n] = x[n]; yp[n] = y[n];
		}
		_fitPoly	= new double [3*nPoints+3*maxDegree+3];
		weights[0]	= -1.;		// do an unweighted least-square-fit
		_rms		= 1.e-20;	// polynomal fit will not achieve this RMS,
								// thus, we will always use the maximum
								// possible degree
		int ierr, degree;		//	error code from SLATEC

        SLATEC_DPOLFT(nPoints, xp, yp, weights, maxDegree, degree, _rms,
					  vals, ierr, _fitPoly);
		_degree = degree;

		if (ierr != 1 && ierr != 3)
			//
			//	fit failed
			//
			errstr<<"Routine `DPOLFIT' failed to find a polynomal fit of "
					"maximum degree "<<maxDegree
				  <<" - return code is "<<ierr
				  <<Sas::error("UnsuccessfulFit");

		//
		//	fit is ok
		//
		errstr<<"Fit polynomial computed - optimized degree: "<<_degree
			  <<" - RMS error: "<<_rms
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		delete [] weights;
		delete [] vals;
		delete [] xp;
		delete [] yp;
	} else
		errstr<<"Only "<<nPoints<<" data points are available - cannot do a "
				"polynomial fit"
			  <<Sas::error("TooFewDataPoints");

	ENDPROC
} // FitPolynomial::FitPolynomial()


double FitPolynomial::evaluate(double x) const
{
	PROC

	double retVal;

	SLATEC_DP1VLU(_degree, 0, x, retVal, static_cast<double *>(0),
				  _fitPoly);

	RETURN retVal;
} // FitPolynomial::evaluate()


double FitPolynomial::derivative(double x, unsigned short n) const
{
	PROC

	if (n==0)
		errstr<<"Derivative order must be greater than or equal to one"
			  <<Sas::error("ArgumentOutOfRange");

	double retVal;
	double deriv[n];
	int di = INT(n);

	SLATEC_DP1VLU(_degree, di, x, retVal, deriv, _fitPoly);

	RETURN deriv[n-1];
} // FitPolynomial::derivative()



FitPolynomial::~FitPolynomial()
{
	PROC
	delete [] _fitPoly;
	ENDPROC
} // FitPolynomial::~FitPolynomial()
