//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrelator	--- user friendly front-end to ODF Time Correlation File
//
//	DESCRIPTION:
//		A class TimeCorrelator is defined which provides methods to perform
//		the following time-related conversions:
//
//			o OBTcoarse/OBTfine -> Mission Ellapsed Time tag
//			o ERT				-> Mission Ellapsed Time tag
//			o MET				-> On-Board Time tag
//
//		These conversions make use of the data in the Spacecraft Time
//		Correlation File which is part of every ODF.
//
//	CHANGE HISTORY:
//		UL, 1998-07-15: 1.0.0: - first version
//		UL, ...
//		UL, 2001-03-02: 1.2.0: - TimeCorrelator construction method can
//								 be chosen via argument
//		UL, 2001-03-30: 1.3.0: - added
//									TimeCorrelator::toOnBoardTime()
//							   - give some routines better names
//
//	USED ACRONYMS:
//		AHF	: Attitude History File
//		ERT	: Earth Received Time
//		OBT	: On-Board-Time
//		ODF	: Obervation Data File
//		TCF	: Time Correlation File
//		RMS	: Root-Mean-Square
//		UTC	: Coordinated Universal Time
//		MET : Mission-Ellapsed-Time (number of elapsed seconds since
//		      fixed mission reference time 1998-01-01T00:00:00 TT)
//
//	AUTHOR:
//		UL, Wed Jul 15 14:38:04 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/TimeCorrelator.h,v 1.8 2001/04/02 11:46:05 ulammers Exp $
///

#ifndef TIME_CORRELATOR_H
#define TIME_CORRELATOR_H

#include "Odf.h"
#include "Xmm.h"
#include "STime.h"
#include "OalChild.h"

#include <vector>

class OalServer;
class FitPolynomial;	// forward declaration

class TimeCorrelator : public virtual OalChild<OalServer> {
public:
	TimeCorrelator(const STime &timeBase = Xmm::missionReferenceTime);
		//
		//	PURPOSE:
		//		Create dummy instance of TimeCorrelator object
		//
		//	ARGUMENTS:
		//		I:	timeBase
		//				the time base all to be returned time tags will refer
		//				to
		//

	typedef struct {
		unsigned	degree;			//	degree of polynomial
		double		rms;			//	RMS - quality of fit
		double		discardPoints;	//	fraction of points to be discared
									//	from data set to achive given RMS
	} FitPolySpecs;

	enum ConstructionMethod { TCS_ONLY = 0, TCS_RECALCULATE_DELAYS };
	TimeCorrelator(const OalServer *oalEngine,
				   const FitPolySpecs &obtMetFitPolySpecs,
				   const FitPolySpecs &ertMetFitPolySpecs,
				   ConstructionMethod cm = TCS_ONLY,
				   const STime &timeBase = Xmm::missionReferenceTime);

	TimeCorrelator(const OdfSet *timeCorrelationSet,
				   const FitPolySpecs &obtMetFitPolySpecs,
				   const FitPolySpecs &ertMetFitPolySpecs,
				   ConstructionMethod cm = TCS_ONLY,
				   const STime &timeBase = Xmm::missionReferenceTime);

	TimeCorrelator(const string &timeCorrelationSetName,
				   const FitPolySpecs &obtMetFitPolySpecs,
				   const FitPolySpecs &ertMetFitPolySpecs,
				   ConstructionMethod cm = TCS_ONLY,
				   const STime &timeBase = Xmm::missionReferenceTime);
		//
		//	PURPOSE:
		//		Create instance of TimeCorrelator object from contents
		//		of ODF TCF; the latter contains the OBT-to-UTC relation in
		//		a tabulated form, i.e. a list of diskrete OBT-UTC data points.
		//		This data will be read and fitted to to a polynomal of a
		//		maximum allowed degree which will be used by the
		//		below methods toMissionEllapsedTime() to perform
		//		conversions from the OBT/ERT to MET. Leap seconds in UTC are
		//		properly treated.
		//
		//	ARGUMENTS:
		//		I:	timeCorrelationSet
		//				must refer to an ODF TCF otherwise an error condition
		//				is raised
		//
		//			timeCorrelationSetName
		//				the full name (dir+basename) of an ODF TCF; an
		//				error condition is raised if the specified file is not
		//				valid
		//
		//			timeBase
		//				the time base all to be returned time tags will refer
		//				to
		//			fitPolySpecs
		//				see description in constructor of FitPolynomial()
		//				below


	const FitPolySpecs &obtMetFitPolySpecs() const;
	const FitPolySpecs &ertMetFitPolySpecs() const;
		//
		//	PURPOSE:
		//		Inquire data of the fit OBT/ERT-UTC fit polynomial;
		//
		//	RETURNS:
		//		structure with the following fields:
		//			<degree>		: actual degree of the fit polynomial
		//							  can be lower or equal to the value
		//							  specified in construtor call
		//			<rms>			: RMS as a measure of the quality
		//							  of the fit
		//			<discardPoints>	: fraction of points discarded from
		//							  original data set in ODF TCF
		//


	void setBaseTime(const STime &newBaseTime) const;
		//
		//	PURPOSE:
		//		Change the base time specified in any of the above
		//		constructors
		//
		//	ARGUMENTS:
		//		I:	newBaseTime
		//				the new base time to be used
		//


	double toMissionEllapsedTime(double obtTag) const;
	double toMissionEllapsedTime(unsigned obtCoarse,
								 unsigned short obtFine) const;
		//
		//	PURPOSE:
		//		Methods to convert a given On-Board-Time (OBT) tag into
		//		a offset time in seconds measured from the base time
		//		specified in the constructor calls or via
		//		TimeCorrelator::setBaseTime(). The OBT-UTC interpolation
		//		curve which has been set up during the object construction
		//		is employed. The quantity returned is directly usable for
		//		the time tag fields in event lists and HK files. The OBT
		//		values are supposed to monotonically increase from call to
		//		call. A decrease with respect to the previous value is
		//		interpreted as a clock wrap-around which is then properly
		//		handled.
		//
		//		Please note: TimeCorrelator::resetObtMetRelation
		//					 below resets the OBT-MET conversion, i.e.,
		//					 a decreasing OBT value is then _not_ considered
		//					 as a clock wrap around
		//
		//		ARGUMENTS:
		//			I:	obtCoarse/obtFine
		//					Coarse and fine component of an OBT tag
		//					as a pair of 32 and 16 bit unsigned values.
		//					The coarse time is a integral second
		//					counter, while the fine time measures the
		//					fractional part of a second with a resolution
		//					of 2^(-16) s = 1/65536 sec.
		//
		//				obttag
		//					A Coarse/Fine time pair combined as a double:
		//						obttag = obtCoarse + obtFine/65536.
		//
		//		RETURNS:
		//			The difference in seconds to the base time (mission
		//			reference time) corresponding to a given OBT value.
		//


	double toOnBoardTime(double missionEllapsedTime) const;
		//
		//	PURPOSE:
		//		The inverse of toMissionEllapsedTime(), i.e., convert
		//		a given MET to an OBT value
		//
		//	ARGUMENTS:
		//		missionEllapsedTime	: MET value [s]
		//
		//	RETURNS:
		//		OBT value corresponding to given MET
		//


	void resetObtMetRelation() const;
		//
		//	PURPOSE:
		//		Reset the OBT-MET relation. A call to this method
		//		between two calls TimeCorrelator::toMissionEllapsedTime() with
		//		a decreasing OBT value will prevent the decrease to be
		//		interpreted as a clock wrap around.
		//


	void plotObtMetRelation() const;
		//
		//	PURPOSE:
		//		Plot the OBT-MET relation using PGPLOT. The
		//		environment variable PGPLOT_DEV determines the
		//		plotting device.
		//


	double toMissionEllapsedTime(const string &ertString) const;
	double toMissionEllapsedTime(const STime &ertTag) const;
		//
		//	PURPOSE:
		//		Same as TimeCorrelator::toMissionEllapsedTime() above
		//		but using an ERT tag instead of OBT. Please note: ERT values
		//		cannot decrease - there is no wrap-around case.
		//
		//	ARGUMENTS:
		//		I:	ertString
		//				ERT tag in the format yyyy-mm-ddThh:mm:ss[.sss]
		//
		//			ertTag
		//				ERT tag as an STime object
		//
		//	RETURNS:
		//		The difference in seconds to the base time (mission
		//		reference time) corresponding to a given ERT value.
		//


	void plotErtMetRelation() const;
		//
		//	PURPOSE:
		//		Plot the ERT-MET relation using PGPLOT. The
		//		environment variable PGPLOT_DEV determines the
		//		plotting device.
		//


	const FitPolynomial *obt2MetPoly() const { return _obtMetPoly; }
	const FitPolynomial *ert2MetPoly() const { return _ertMetPoly; }
	const FitPolynomial *met2ObtPoly() const { return _metObtPoly; }
		//
		//	PURPOSE:
		//		Give direct access to the OBT/ERTtoMET and METtoOBT
		//		fit polynomials
		//


	~TimeCorrelator();
		//
		//	PURPOSE:
		//		Destroy TimeCorrelator object constructed with any of the
		//		above constructors
		//

private:
	//	methods
	void construct(const string &, ConstructionMethod);
	void constructFromTCSonly(const string &);
	void constructFromTCSrecalcDelays(const string &);

	const FitPolynomial *doPolyFit(FitPolySpecs &, vector<double> &,
								   vector<double> &) const;

	//	data members
	FitPolySpecs			_obtMetFitPolySpecs, _ertMetFitPolySpecs;
	mutable STime			_baseTime;
	STime					_utcBaseTime;
	double					_utcOffset;
	mutable double			_lastObtTag;
	mutable unsigned		_obtWraps;
	const FitPolynomial		*_obtMetPoly, *_ertMetPoly, *_metObtPoly;
};


class FitPolynomial {
public:
	FitPolynomial() {}

	FitPolynomial(const vector<double> &x, const vector<double> &y,
				  const unsigned maxDegree = 1, double rms = 1.e-10);
		//
		//	PURPOSE:
		//		Construct from a set of discrete data points a fit polynomial,
		//		of a maximum degree	with SLATEC-routine `DPOLFT'. This method
		//		automatically optimizes the degree of the polynomial based on
		//		a given RMS to achieved. If no satisfactory fit can be
		//		attained, an error condition will be raised
		//
		//	ARGUMENTS:
		//		I:	x
		//				x values ...
		//			y
		//				... and corresponding ys
		//			maxDegree/rms
		//				specifications controlling the polynomial fit of the
		//				data; <maxDegree>
		//				defines the maximum allowed degree (e.g. 1 == linear)
		//				of the fit polynomal; <rms> is the to be obtained RMS
		//				value; it is possible that no polynomial with a degree
		//				lower than or equal to <polyDegree> can be found which
		//				satisfies the <rms> criterion. In this case it is
		//				attempted to achive the desired result by successively
		//				removing individial points from the data
		//				set: The respective point which is most "off" the
		//				computed fit polynomal will be removed first, the fit
		//				repeated, etc. till the requested RMS has been
		//				achieved or the fraction of removed points exceeds
		//				<discardPoints>. In the latter case a fatal error
		//				condition will be raised.
		//


	double evaluate(double x) const;
	double derivative(double x, unsigned short n=1) const;
		//
		//	PURPOSE:
		//		Evaluate fit polynomial/derivative of fit polynomial at a
		//		specific point
		//
		//	ARGUMENTS:
		//		value	: point at which polynomial is to be evaluated
		//		n		: derivative to evaluate (>=1)
		//
		//	RETURNS:
		//		value of polynomial/derivative at <x>
		//


	unsigned degree() const { return _degree; }
	double rms() const { return _rms; }
		//
		//	PURPOSE:
		//		Return actual degree and achieved RMS of the fit polynomal;
		//		the values can deviate from the specifications given
		//		in the constructor call
		//


	const vector<double> xPoints() const { return _x; }
	const vector<double> yPoints() const { return _y; }
		//
		//	PURPOSE:
		//		Return input data points with which the object was constructed
		//


	~FitPolynomial();
		//
		//	PRUPOSE:
		//		destroy object constructed with any of the above constructors
		//

private:
	unsigned	_degree;	//	degree of fit polynomial
	double		_rms;		//	RMS - quality of fit
	double		*_fitPoly;	// the result of the fit is a populated double
							// vector defining the polynomal - will be used
							// by evaluate() to compute polynomial at
							// a certain point
	vector<double>	_x, _y;	// input data points
};

#endif
