//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeTags.cc	--- implementation of TimeTag classes
//
//	DESCRIPTION:
//		Implementation part of classes in Tag.h
//
//	CHANGE HISTORY:
//		UL, 1998-08-03: 1.0.0: - first version
//		UL, 1998-10-14: 1.1.0: - bug fix in toObt() methods
//		UL, 1999-03-04: 1.2.0: - changes to make it work with low memort Dal
//		UL, 1999-03-16: 1.3.0: - changed column name: "EOSCOARSE" -> "EOSCOARS"
//		UL, 1999-04-28: 1.4.0: - fatal -> error
//		UL, 1999-05-25: 1.4.1: - bug fix: was assuming wrong range of
//								 coarse times for all instruments
//							   - frameCounterToObt: do not bail out on
//								 missing frame numbers; just issue
//								 warning message and set OBT to -1.
//							   - modification in frameCounterToObt for EMOS
//		UL, 1999-11-02: 1.4.2: - more informative error message in
//								 frameCounter->obt conversion for RGS
//		UL, 1999-12-17: 1.5.0: - bug fix in
//									InstrFrameCounterVector<EPNcam>::toObt()
//								 was accessing wrong AUX file
//		UL, 1999-12-20: 1.6.0: - work-around for SciSim bug: FTCOARSE for
//								 MOS/PN AUX files in SciSim generated
//								 AUX files is not in range 0..32767 but
//								 0..32767,-32768,-32767...
//		UL, 2000-01-05: 1.7.0: - bug fix in resetTimeEpic(),
//								 fixes:
//								 SPR 957 : Problem with OAL and diagnostic
//										   files on SVT3
//								 SPR 1063 : Behaviour when exposure is not 1.
//		UL, 2000-01-11: 1.8.0: - brought in line with ODF ICD issue 2.4
//								 (FTFINE columns in PN/RGS AUX files are
//								  now unsigned)
//		UL, 2000-02-03: 1.9.0: - fixed SPR  1079 : Unwarranted IllegalObtValue
//								 warning
//								 (emframes subtracts from the OBT value of
//								 the first frame the frame integration time
//								 to estimate the start of the frame. If
//								 the EPIC clock reset time in the PEH
//								 (FTCOARSE/FTFINE) is close to zero this can
//								 then lead to negative OBT times which was
//								 checked against in ObtVector::toUtc(); in
//								 this case the method used to set the time
//								 tag to -1 and issue a warning message; now
//								 it does merely do the latter but computes the
//								 time tag ok).
//		UL, 2000-04-03: 1.10.0:- accessing coarse/fine time components of
//								 EPIC clock reset times in periodic HK
//								 files with right names:
//
//										coarse components	fine component
//											 name/type			 name/type
//								 EMOS1		E1573/J(S)			EU573/I(U)
//								 EMOS2		K1573/J(S)			KU573/I(U)
//								 EPN		F1030/J(S)			FU036/I(U)
//							   - bug fix in frame counter to tag conversions
//								 which falsely caused a wrong-wrap around
//								 detected to be detected
//								 (this fixes SPR 1239 : Wrong wrap around at
//								  first frame)
//		UL, 2000-08-29: 1.11.0:- various changes to read coarse time wrap
//								 around value for EPN clock from periodic
//								 HK parameter F1052 (was assumed to be
//								 constant=32768)
//		UL, 2001-03-27: 1.11.1:- open all HK/AUX data files in lowmem mode
//							   - check the validity of the EPIC clock reset
//								 time/OBT values converted from RGS/OM time
//								 tags by comparison against exposure end
//								 time; if the former lies after the latter
//								 there was a CDMU crash - take the exposure
//								 start time as effective reset time in this
//								 case
//
//	AUTHOR:
//		U. Lammers, Mon Aug  3 16:10:16 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/TimeTags.cc,v 1.1 2001/04/02 11:46:05 ulammers Exp $
//

#include <cassert>

#include "Dal.h"
#include "Tracer.h"
#include "TimeTags.h"
#include "Xmm.h"
#include "Odf.h"
#include "OalServer.h"
#include "CalOalUtility.h"
#include "OalDalUtils.h"

using namespace CalOalUtility;

static const double ObtFineTimeResolution = 1./65536.;
	//	16-bit fine time component of Central On-Board Clock will tick
	//	at maximum resolution: 2^-16 sec ~ 15.3 usec


class PeriodicHKParamFinder {
public:
	PeriodicHKParamFinder(const OalServer *server) :
		_server(server), _table(0) {
		PROC

		_instr = _server->instrument();
		const OdfSet *hkSet;
		if (_instr == Xmm::EMOS1 || _instr == Xmm::EMOS2)
			hkSet = server->odf()->selectSet(Xmm::PERIODIC_HOUSEKEEPING);
		else
			//	for PN the reset time is in the main periodic HK
			hkSet = server->odf()
								->selectSet(Xmm::MAIN_PERIODIC_HOUSEKEEPING);

		assert(hkSet);

		errstr<<"Opening periodic HK set `"
			  <<hkSet->name()<<"' to look for needed parameters"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		_table = dataSetServer->open(hkSet->name(), DataSet::Read,
									 DataSetServer::Low)->table(0);
		_index = 0;

		ENDPROC
	} // PeriodicHKParamFinder()


	void seek(const STime &t) const {
		PROC

		double tt = t - Xmm::missionReferenceTime + 30.;
		errstr<<"looking for periodic HK parameter at t="<<tt<<endl
			   <<Sas::message(Msg::Lib, Msg::Verbose);

		const TimeCorrelator *timeCorrelator = _server->timeCorrelator();

		ColumnData *timeCol = _table->column("TIME")->data();

		unsigned nRows = _table->rows(), n;
		for (n=0; n<nRows; ++n) {
			string time(stringColumnRow(timeCol, n));
			double sec = timeCorrelator->toMissionEllapsedTime(time);
			errstr<<time<<" =^ "<<sec<<endl
				  <<Sas::message(Msg::Lib, Msg::Verbose);
			if (sec >= tt)
				//	found row
				break;
		}
		delete timeCol;

		if (n >= nRows)
			errstr<<"Could not find time t="<<Xmm::missionReferenceTime+tt
				  <<" in periodic HK table `"
				  <<_table->name()<<"'"<<Sas::error("TimeMismatch");

		_index = n;

		ENDPROC
	} // seek()


	//
	//	PURPOSE:
	//		Get the value of the EPIC (MOS or PN) clock from the corresponding
	//		HK file.
	//
	//		The EPIC clock is reset once at the beginning of every
	//		exposure. The time of this event is available as a COBT stamp
	//		(Coarse+Fine Time) in the periodic HK file belonging to this
	//		exposure. The function reads the value of reset time corresponding
	//		to 30 sec after the start of the exposure.
	//
	//	RETURNS:
	//		EPIC clock reset time of current exposure as a single double COBT
	//		value
	//
	double resetTime() const {
		PROC

		string ftCoarseName,
			   ftFineName;
		switch (_instr) {
			case Xmm::EMOS1: ftCoarseName	= "E1573";
							 ftFineName		= "EU573";
							 break;
			case Xmm::EMOS2: ftCoarseName	= "K1573";
							 ftFineName		= "KU573";
							 break;
			case Xmm::EPN:	 ftCoarseName	= "F1030";
							 ftFineName		= "FU036";
							 break;
			default:		 break;
		}

		ColumnData *cdc = _table->column(ftCoarseName)->data(),
				   *cdf = _table->column(ftFineName)->data();
		int32 *coarseTime = cdc->int32Data();
		uint16 *fineTime = cdf->uint16Data();

		double resetTime = DBL(OalDalUtils::unsigned32(coarseTime[_index])) +
								ObtFineTimeResolution *
							DBL(OalDalUtils::unsigned32(fineTime[_index]));

		//
		//	check that the reset time is not later than the exposure end
		//	time; if this IS the case there was probably a CDMU crash
		//	during the exposure and the reset time is wrong, i.e., shows
		//	the OBT value before the crash; if we continue with this
		//	wrong (too large) value all MET tags calculated from the TCS shall
		//	be wrong as well. Examples are all ODFs from revolutions
		//	103+104 (also F. Gianinni's email on CDMU crashes dated
		//	2001-03-29)
		//
		//	In this case we carry on with the OBT value corresponding
		//	to the actual exposure start - Note: This value can be
		//	off by a couple of seconds, i.e., all METs will be afflicted
		//	with the same absolute error.
		//
		STime tstart(_server->odf()->currentExposure()->actualStartTime()),
			  tend(_server->odf()->currentExposure()->actualEndTime());

		const TimeCorrelator *timeCorrelator = _server->timeCorrelator();
		STime treset(Xmm::missionReferenceTime +
					 timeCorrelator->toMissionEllapsedTime(resetTime));
		timeCorrelator->resetObtMetRelation();
		double deltaT = tend-treset;
		if (deltaT<0.) {
			errstr<<"The "<<Xmm::toString(_instr)<<" clock reset time read "
					"from HK parameters "<<ftCoarseName<<'/'<<ftFineName
				  <<" lies "<<-deltaT<<"s after the end of the current "
					"exposure at "<<tend<<". The most likely cause of this is "
					"a CDMU crash which occured during the exposure. Will "
					"set the reset time to the exposure start at "
				  <<tstart<<". Please note: This value can be inaccurate "
					"by several seconds and thus, all event time tags shall "
					"be afflicted with the same absolute error."
				  <<Sas::warning("WrongClockResetTime");
			resetTime = timeCorrelator->toOnBoardTime(tstart -
												Xmm::missionReferenceTime);
		}

		errstr<<"EPIC clock reset time: "<<resetTime<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		delete cdc;
		delete cdf;

		RETURN resetTime;
	} // resetTime()


	//
	//	PURPOSE:
	//
	//	RETURNS:
	//
	double wrapValue() const {
		PROC

		if (_instr != Xmm::EPN)
			ALTRETURNV(32768.);

		if (!_table->hasColumn("F1052")) {
			errstr<<"Expected Periodic HK parameter F1052 missing from `"
				  <<_table->name()<<"' - will continue with nominal value "
					"32400 as coarse time wrap value for EPN"
				  <<Sas::warning("NoSuchPeriodicHKparam");
			ALTRETURNV(32400.);
		}

		ColumnData *cdf1052 = _table->column("F1052")->data();
		double val = cdf1052->uint16Data()[_index];
		delete cdf1052;

		RETURN val;
	} // wrapValue()


	~PeriodicHKParamFinder() {
		PROC
		if (_table)
			dataSetServer->close(_table->parent());
		ENDPROC
	} // ~PeriodicHKParamFinder()

private:
	const OalServer	*	_server;
	Xmm::Instrument		_instr;
	mutable unsigned	_index;
	Table *				_table;
};

//
//	---( EMOS )-------------------------------------------------------------
//
unsigned InstrFrameCounterVector<EMOScam>::_nWraps = 0;
double InstrFrameCounterVector<EMOScam>::_resetTime = -1.;
double InstrFrameCounterVector<EMOScam>::_wrapValue;

void InstrFrameCounterVector<EMOScam>::setStatics() const
{
	PROC

	if (_resetTime < 0.) {
		PeriodicHKParamFinder pf(&oalServer);
		STime tstart(parent()->odf()->currentExposure()->actualStartTime());
		pf.seek(tstart);

		_resetTime = pf.resetTime();
		_wrapValue = pf.wrapValue();
	}

	ENDPROC
} // setStatics()


static const double EmosFineTimeResolution = 40.e-6;	//	EPIC-EST-SP-001,
														//	page 33

ObtVector InstrFrameCounterVector<EMOScam>::toObt() const
{
	PROC

	setStatics();	//	make sure the EPIC reset time/wrap value has been read
					//	from HK

	//
	//	here we need to look in the EMOS Auxiliary File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::AUXILIARY);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdccdId = auxTable->column("CCDID")->data(),
			   *cdccdNode = auxTable->column("CCDNODE")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32 *frame = cdframe->int32Data();
	int8 *ccdId = cdccdId->int8Data(),
		 *ccdNode = cdccdNode->int8Data();
	int16 *ftcoarse = cdftcoarse->int16Data(),
		  *ftfine = cdftfine->int16Data();

	int8 ccd = parent()->ccd(),
		 node = parent()->node();

	ObtVector obtVector(this);
	unsigned nCounter = size();
	obtVector.reserve(nCounter);
	unsigned lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();
	for (unsigned n=0; n<nCounter; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i) {
			if (ccdId[i] == ccd && ccdNode[i] == node &&
				frame[i] == (*this)[n]) {
				//	found it

				unsigned coarseTime = OalDalUtils::unsigned32(ftcoarse[i]);
				//
				//	check for wrap-arounds
				//
				if (coarseTime < lastCoarseTime)
					++_nWraps;

				lastCoarseTime = coarseTime;
				obtVector.push_back(
						   DBL(_nWraps) * _wrapValue + DBL(coarseTime) +
						   DBL(OalDalUtils::unsigned32(ftfine[i]))
									* EmosFineTimeResolution +
						   _resetTime);
							//	FTCOARSE is 15 bit - range 0..32767
				indx = i;
				break;
			}
		}

		if (i >= nRecs) {
			//
			//	we have not found the time of a frame counter
			//	let's bail out
			errstr<<"Could not find frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" for CCD/node: "<<UNT(ccd)<<"/"<<UNT(node)
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to -1"
				  <<Sas::warning("MissingFrameNumber");
			obtVector.push_back(-1.);
			indx = 0;
		}
	}


	delete cdframe;
	delete cdccdId;
	delete cdccdNode;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	RETURN obtVector;
} // InstrFrameCounterVector<EMOScam>::toObt()



//
//	---( EPN )-------------------------------------------------------------
//
unsigned InstrFrameCounterVector<EPNcam>::_nWraps = 0;
double InstrFrameCounterVector<EPNcam>::_resetTime = -1.;
double InstrFrameCounterVector<EPNcam>::_wrapValue;

void InstrFrameCounterVector<EPNcam>::setStatics() const
{
	PROC

	if (_resetTime < 0.) {
		PeriodicHKParamFinder pf(&oalServer);
		STime tstart(parent()->odf()->currentExposure()->actualStartTime());
		pf.seek(tstart);

		_resetTime = pf.resetTime();
		_wrapValue = pf.wrapValue();
	}

	ENDPROC
} // InstrFrameCounterVector<EPNcam>::setStatics()

static const double EpnFineTimeResolution = 20.48e-6;	//	EPIC-EST-SP-001,
														//	page 33

ObtVector InstrFrameCounterVector<EPNcam>::toObt() const
{
	PROC

	setStatics();	//	make sure the EPIC reset time has been
						//	read from HK

	//
	//	here we need to look in the EPN Auxiliary File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::AUXILIARY);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdccdId = auxTable->column("CCDID")->data(),
			   *cdquadId = auxTable->column("QUADRANT")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32 *frame = cdframe->int32Data();
	int8 *ccdId = cdccdId->int8Data(),
		 *quadId = cdquadId->int8Data();
	int16 *ftcoarse = cdftcoarse->int16Data();
	uint16 *ftfine = cdftfine->uint16Data();

	int8 ccd = EPNcam::toCcdId(parent()->ccd()),
		 quad = EPNcam::toQuadrantId(parent()->ccd());

	ObtVector obtVector(this);
	unsigned nCounter = size();
	obtVector.reserve(nCounter);
	unsigned lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();

	for (unsigned n=0; n<nCounter; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i)
			if (ccdId[i] == ccd && quadId[i] == quad &&
				frame[i] == (*this)[n]) {
				//	found it

				unsigned coarseTime = OalDalUtils::unsigned32(ftcoarse[i]);
				//
				//	check for wrap-arounds
				//
				if (coarseTime < lastCoarseTime)
					++_nWraps;

				lastCoarseTime = coarseTime;
				obtVector.push_back(
							   DBL(_nWraps) * _wrapValue + DBL(coarseTime) +
							   DBL(OalDalUtils::unsigned32(ftfine[i]))
										* EpnFineTimeResolution +
						   _resetTime);
							//	FTCOARSE is 15 bit: 0..32767
				indx = i;
				break;
			}

		if (i >= nRecs) {
			// 
			//	we have not found the time of a frame counter
			//	let's bail out
			//
			errstr<<"Could not find frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" for CCD/Quadrant: "<<UNT(ccd)<<"/"<<UNT(quad)
				  <<" ["<<parent()->ccd()<<"]"
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to -1"
				  <<Sas::error("MissingFrameNumber");
			obtVector.push_back(-1.);
			indx = 0;
		}

	}

	delete cdframe;
	delete cdccdId;
	delete cdquadId;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	RETURN obtVector;
} // InstrFrameCounterVector<EPNcam>::toObt()


//
//	---( RGS )--------------------------------------------------------------
//
//
//	other than EMOS and EPN, RGS and OM use the COBT values directly
//	to time stamp frame readout times - no need to read a reset value
//
void InstrFrameCounterVector<RGScam>::setStatics() const
{
	PROC
	_resetTime = 0.;
	ENDPROC
}

unsigned InstrFrameCounterVector<RGScam>::_nWraps = 0;
double InstrFrameCounterVector<RGScam>::_resetTime = -1.;

static const double RgsFineTimeResolution = ObtFineTimeResolution;

ObtVector InstrFrameCounterVector<RGScam>::toObt() const
{
	PROC

	setStatics();	//	set reset time to 0.

	//
	//	here we need to look in the RGS Auxiliary File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::AUXILIARY);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdccdId = auxTable->column("CCDID")->data(),
			   *cdftcoarse = auxTable->column("EOSCOARS")->data(),
			   *cdftfine = auxTable->column("EOSFINE")->data();
	int32 *frame = cdframe->int32Data();
	int8 *ccdId = cdccdId->int8Data();
	int32 *ftcoarse = cdftcoarse->int32Data();
	uint16 *ftfine = cdftfine->uint16Data();

	int8 ccd = parent()->ccd();

	ObtVector obtVector(this);
	unsigned nCounter = size();
	obtVector.reserve(nCounter);
	unsigned lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();
	for (unsigned n=0; n<nCounter; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i)
			if (ccdId[i] == ccd && frame[i] == (*this)[n]) {
				//	found it

				unsigned coarseTime = OalDalUtils::unsigned32(ftcoarse[i]);
				//
				//	check for wrap-arounds
				//
				if (coarseTime < lastCoarseTime)
					++_nWraps;

				lastCoarseTime = coarseTime;
				obtVector.push_back(
						   DBL(_nWraps) * 65536. * 65536. +
						   DBL(coarseTime) +
						   DBL(OalDalUtils::unsigned32(ftfine[i]))
									* RgsFineTimeResolution +
						   _resetTime);
				indx = i;
				break;
			}
		if (i >= nRecs) {
			//
			//	we have not found the time of a frame counter
			//	let's bail out
			//
			errstr<<"Could not find frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" for CCD: "<<UNT(ccd)
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to -1"
				  <<Sas::warning("MissingFrameNumber");
			obtVector.push_back(-1.);
			indx = 0;
		}

	}

	//	tidy up
	delete cdframe;
	delete cdccdId;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	//
	//	check that OBT values lie before end of exposure; if not we had
	//	a CDMU crash (see comment above for PeriodicHKParamFinder::resetTime())
	//	and need to correct all values
	//
	if (obtVector.size() > 0) {
		STime tstart(parent()->odf()->currentExposure()->actualStartTime()),
			  tend(parent()->odf()->currentExposure()->actualEndTime());
		const TimeCorrelator *timeCorrelator = parent()->timeCorrelator();
		STime tfirst(Xmm::missionReferenceTime +
					 timeCorrelator->toMissionEllapsedTime(obtVector[0]));
		timeCorrelator->resetObtMetRelation();
		double deltaT = tend-tfirst,
			   corrT = 0.;
		if (deltaT<0.) {
			errstr<<"The On-Board-Time values corresponding to the RGS "
					"frame counters lie after the end of the current exposure "
					"at "<<tend<<". The most likely cause of this is a "
					"CDMU crash that occured during the exposure. Will "
					"adjust all times such that the time of the first frame "
					"coincides with the start of the the exposure at "
				  <<tstart<<". Please note: This value can be inaccurate "
					"by several seconds and thus, all event time tags shall "
					"be afflicted with the same absolute error."
				  <<Sas::warning("PossiblyInaccurateOBTValues");
			corrT = timeCorrelator->toOnBoardTime(tstart -
												Xmm::missionReferenceTime) -
					obtVector[0];
		}
		unsigned N = obtVector.size();
		for (unsigned n=0; n<N; obtVector[n++] += corrT);
	}

	RETURN obtVector;
} // InstrFrameCounterVector<RGScam>::toObt()



//
//	---( OM )--------------------------------------------------------------
//
//
//	other than EMOS and EPN, RGS and OM use the COBT values directly
//	to time stamp frame readout times - no need to read a reset value
//
void InstrFrameCounterVector<OMcam>::setStatics() const
{
	PROC
	_resetTime = 0.;
	ENDPROC
}

unsigned InstrFrameCounterVector<OMcam>::_nWraps = 0;
double InstrFrameCounterVector<OMcam>::_resetTime = -1.;

static const double OmFineTimeResolution = 9.765625E-4; // ODF ICD p. 73,
														// bullet 3

ObtVector InstrFrameCounterVector<OMcam>::toObt() const
{
	PROC

	setStatics();	//	set reset time to 0.

	Xmm::DataMode currentMode = parent()->dataMode();
	if (currentMode != Xmm::DATA_FAST)
		errstr<<"Converting "<<Xmm::toString(Xmm::OM)<<" frame counters "
				"makes only sense in "<<Xmm::toString(Xmm::FAST)<<" mode"
				"current mode is: `"<<Xmm::toString(currentMode)<<"'\n"
			  <<Sas::error("IllegalMode");

	//
	//	here we need to look in the OM Fast Mode Event File
	//
	const OdfSet *odfSet = parent()->odf()->selectSet(Xmm::DATA_FAST);
	assert(odfSet);
	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
										  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32 *frame = cdframe->int32Data(),
		  *ftcoarse = cdftcoarse->int32Data();
	int16 *ftfine = cdftfine->int16Data();

	ObtVector obtVector(this);
	unsigned nCounter = size();
	obtVector.reserve(nCounter);
	int16 lastCoarseTime = 0;

	unsigned indx = 0,
			 nRecs = auxTable->rows();

	for (unsigned n=0; n<nCounter; ++n) {
		unsigned i;
		for (i=indx; i<nRecs; ++i)
			if (frame[i] == (*this)[n]) {
				//	found it
				int16 coarseTime = ftcoarse[i];
				//
				//	check for wrap-arounds
				//
				lastCoarseTime = coarseTime;
				if (coarseTime < lastCoarseTime)
					++_nWraps;
				obtVector.push_back(
						   DBL(_nWraps) * 16384. +
						   DBL(coarseTime) +
						   DBL(OalDalUtils::unsigned32(ftfine[i]))
									* OmFineTimeResolution +
						   _resetTime);
				indx = i;
				break;
			}

		if (i >= nRecs) {
			//
			//	we have not found the time of a frame counter
			//	let's bail out
			errstr<<"Could not find time for frame number "
				  <<OalDalUtils::unsigned32((*this)[n])
				  <<" in auxiliary file `"<<auxTable->qualifiedName()<<"'"
				  <<" - will set OBT to -1"
				  <<Sas::error("MissingFrameNumber");
			obtVector.push_back(-1.);
			indx = 0;
		}

	}

	delete cdframe;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	//
	//	check that OBT values lie before end of exposure; if not we had
	//	a CDMU crash (see comment above for PeriodicHKParamFinder::resetTime())
	//	and need to correct all values
	//
	if (obtVector.size() > 0) {
		STime tstart(parent()->odf()->currentExposure()->actualStartTime()),
			  tend(parent()->odf()->currentExposure()->actualEndTime());
		const TimeCorrelator *timeCorrelator = parent()->timeCorrelator();
		STime tfirst(Xmm::missionReferenceTime +
					 timeCorrelator->toMissionEllapsedTime(obtVector[0]));
		timeCorrelator->resetObtMetRelation();
		double deltaT = tend-tfirst,
			   corrT = 0.;
		if (deltaT<0.) {
			errstr<<"The On-Board-Time values corresponding to the OM "
					"time tags lie after the end of the current exposure "
					"at "<<tend<<". The most likely cause of this is a "
					"CDMU crash that occured during the exposure. Will "
					"adjust all times such that the time of the first event "
					"coincides with the start of the the exposure at "
				  <<tstart<<". Please note: This value can be inaccurate "
					"by several seconds and thus, all event time tags shall "
					"be afflicted with the same absolute error."
				  <<Sas::warning("PossiblyInaccurateOBTValues");
			corrT = timeCorrelator->toOnBoardTime(tstart -
												Xmm::missionReferenceTime) -
					obtVector[0];
		}
		unsigned N = obtVector.size();
		for (unsigned n=0; n<N; obtVector[n++] += corrT);
	}

	RETURN obtVector;
} // InstrFrameCounterVector<OMcam>::toObt()


//
//	---( ObtVector )--------------------------------------------------------
//
vector<double> ObtVector::toMET() const
{
	PROC

	unsigned nmax = size();
	vector<double> tags(nmax);

	const TimeCorrelator *timeCorrelator =parent()->parent()->timeCorrelator();

	for (unsigned n=0; n<nmax; ++n) {
		tags[n] = timeCorrelator->toMissionEllapsedTime((*this)[n]);
		if ((*this)[n] < 0.)
			errstr<<"OBT vector element #"<<n<<" is negative ("
				  <<(*this)[n]<<")"
				  <<Sas::warning("NegativeObtValue");
	}

	RETURN tags;
} // ObtVector::toMET()
